from django.core.exceptions import ImproperlyConfigured
from django.utils.importlib import import_module
from django.conf import settings

import os


def load_backend(backend_name):
    try:
        return import_module('.base', 'trackable.message.backends.%s' % backend_name)
    except ImportError, e:
        try:
            return import_module('.base', backend_name)
        except ImportError, e_user:
            backend_dir = os.path.join(os.path.dirname(__file__), 'backends')
            try:
                available_backends = [f for f in os.listdir(backend_dir)
                                      if os.path.isdir(os.path.join(backend_dir, f))
                                      and not f.startswith('.')]
            except EnvironmentError:
                available_backends = []
            available_backends.sort()
            if backend_name not in available_backends:
                error_msg = ("%r isn't an available database backend. \n" +
                    "Try using trackable.message.backends.XXX, where XXX is one of:\n    %s\n" +
                    "Error was: %s") % \
                    (backend_name, ", ".join(map(repr, available_backends)), e_user)
                raise ImproperlyConfigured(error_msg)
            else:
                raise



from django.contrib.contenttypes.models import ContentType
from django.conf import settings

from trackable import TrackableError
from trackable.models import Spider
from trackable.sites import site

import datetime
import warnings
import logging

try:
    import cPickle as pickle
except ImportError:
    import pickle

USER_AGENT_FILTERING = getattr(settings,'TRACKABLE_USER_AGENT_FILTERING', False)
REMOVE_MALFORMED_MESSAGES = getattr(settings,'TRACKABLE_REMOVE_MALFORMED_MESSAGES', False)
CAPTURE_CONNECTION_ERRORS = getattr(settings,'TRACKABLE_CAPTURE_CONNECTION_ERRORS', False)
PROCESS_NUM_MESSAGES = getattr(settings,'TRACKABLE_PROCESS_NUM_MESSAGES', None)
LOGLEVEL = getattr(settings,'TRACKABLE_LOGLEVEL', logging.WARNING)
MAGIC_FLAG = getattr(settings,'TRACKABLE_MAGIC_FLAG', '7R4CK4813M491C')

UNDEFINED_AGENT = 'UNDEFINED_HTTP_USER_AGENT'


class BaseMessageBackend(object):
    """ """
    def send(self, payload):
        """
        Publish ``payload`` to the message backend.

        """
        raise NotImplementedError

    def recv(self):
        """
        Fetch a message from the message backend.

        """
        raise NotImplementedError

    def send_message(self, request, obj, field_name, op_name, data_cls, options={}):
        """
        

        """
        if not data_cls:
            data_cls = site._registry[obj.__class__][0]

        value = None
        try:
            value = options.pop('value')
        except KeyError:
            pass

        message_obj = {}
        message_obj['magic'] = MAGIC_FLAG
        message_obj['data_object_pk'] = data_cls.objects.get_data_object(obj,options).pk
        message_obj['data_cls'] = data_cls
        message_obj['user_agent'] = request.META['HTTP_USER_AGENT'] \
            if 'HTTP_USER_AGENT' in request.META else UNDEFINED_AGENT
        message_obj['field_name'] = field_name
        message_obj['op_name'] = op_name
        message_obj['options'] = options
        message_obj['value'] = value

        # message_body = pickle.dumps( message_obj )
        # self.send(message_body)
        self.send( message_obj )

    def process_messages(self, logger=None, model_cls=None, max_messages=PROCESS_NUM_MESSAGES):
        """
        Process all currently gathered messages by compiling and 
        saving them to the database.
        
        """
        test_condition = lambda x: x < max_messages
        if not max_messages:
            test_condition = lambda x: True

        values_lookup = {}
        messages_lookup = {}
        cnt = 0

        if not logger:
            logger = logging.getLogger('trackable')
            console = logging.StreamHandler()
            logger.setLevel(LOGLEVEL)
            logger.addHandler(console)

        while test_condition:
            message = self.recv()
            if not message:
                break

            # message_obj = pickle.loads(message.body)
            message_obj = message.payload
            if 'magic' not in message_obj or message_obj['magic'] != MAGIC_FLAG:
                logger.warning( "Magic bytes not found or do not match. Skipping message." )
                continue

            logger.info( "%d Got message: %s" % (cnt+1,message_obj) )
            cnt += 1

            if USER_AGENT_FILTERING:
                if message_obj['user_agent'] == UNDEFINED_AGENT:
                    msg = "Cannot match: user agent does not exist."
                    logger.warning( msg )
                    message.ack()
                    continue
                hits = Spider.objects.filter( \
                    user_agent__icontains=message_obj['user_agent'][:128])
                if hits:
                    msg = "Not processing potential spider-generated tracking message. User agent=%s" % (message_obj['user_agent'])
                    logger.warning( msg )
                    message.ack()
                    continue

            (op_name,field_name,data_cls,data_object_pk,result) = \
                message_obj['op_name'], message_obj['field_name'], \
                message_obj['data_cls'], message_obj['data_object_pk'], \
                message_obj['value']

            if model_cls and data_cls != model_cls:
                msg = u"Skipping TrackableData type=%s" % (str(data_cls))
                logger.info( msg )
                continue

            data_object = data_cls.objects.get(pk=data_object_pk)

            try:
                getattr(data_object,op_name)
            except AttributeError:
                raise TrackableError( \
                    u"%s does not support %s operation." \
                        % (data_object,op_name))

            op_func = getattr(data_object,op_name)
            op_func(field_name,result)

            data_object = data_cls.objects.get(pk=data_object_pk)
            logger.info( "(%s)->%s on %s: %s" % (data_object,op_name,field_name,result))

            # Acknowledge the message now that the operation has been registered
            message.ack()

        return cnt



from django.views.generic.list_detail import object_detail
from django.shortcuts import get_object_or_404

from trackable.message.counting import send_increment_message
from trackable.tests.models import Page


def page_detail(request, slug):
    response = object_detail(request,Page.objects.all(),slug=slug)
    if response.status_code == 200:
        send_increment_message(
            request,Page.objects.get(slug=slug),'views'
            )
    return response



from django.core.management.base import BaseCommand, CommandError
from django.contrib.contenttypes.models import ContentType
from django.db.models.loading import get_model
from django.db.models import Sum, Avg

from trackable.messaging import process_messages
from trackable.sites import site, NotRegistered

from optparse import make_option

try:
    import cStringIO as StringIO
except ImportError:
    import StringIO

import datetime
import warnings
import codecs
import csv


class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode("utf-8")

class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = StringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        # self.writer.writerow([s.encode("utf-8") for s in row])
        newrow = []
        for s in row:
            if type(s) == unicode:
                s = s.encode("utf-8")
            newrow.append( s )
        self.writer.writerow( newrow )

        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

class Command(BaseCommand):
    help = 'Used to generate CSV summary data for Vendor tracking data.'
    args = 'app.model [app.trackable_model ...]'
    option_list = BaseCommand.option_list + (
        make_option('--aggregate-fields', action='store', dest='aggregate_fields',
                    help='Specify the field on which to aggregate data'),
        )

    def handle(self, model_spec, *trackable_model_specs, **options):
        print "Aggregating data for %s" % (model_spec)
        for option in options.get('aggregate_fields').split(','):
            print "  %s" % (option)

        try:
            app_label,model = model_spec.split('.')
            model_cls = get_model(app_label,model)
            if model_cls is None:
                raise CommandError("Unknown model: %s" % (model_spec))
        except ValueError:
            raise CommandError("Use full appname.ModelName specification for argument %s" % model_spec)

        trackable_models = []
        if trackable_model_specs:
            for model_spec in trackable_model_specs:
                app_label,model = model_spec.split('.')
                model_cls = get_model(app_label,model)
                if model_cls is None:
                    warnings.warn( "Could not resolve trackable model specification: %s" % model_spec )
                trackable_models.append( model_cls )
        try:
            if not trackable_models:
                trackable_models = site._registry[model_cls]
        except KeyError:
            raise CommandError("%s has no registered trackable objects.")

        for trackable_model in trackable_models:
            if trackable_model not in site._registry[model_cls]:
                raise CommandError( \
                    "%s is not a registered trackable object with %s" % \
                        (trackable_model,model_cls))

        fields = [Sum(field_name) for field_name in options.get('aggregate_fields').split(',')]
        fields.extend( [Avg(field_name) for field_name in options.get('aggregate_fields').split(',')] )
        instances = []

        for instance in model_cls.objects.values('id','name'):
            for trackable_model in trackable_models:
                trackable_model_objects = \
                    trackable_model.objects \
                    .filter( \
                        content_type=ContentType.objects.get_for_model(model_cls), \
                        object_id=instance['id'] \
                        )
                if not trackable_model_objects.count():
                    continue
                attrs = trackable_model_objects \
                    .aggregate( \
                        *[field for field in fields] \
                        )
                adjusted_attrs = {}
                for key,val in attrs.iteritems():
                    adjusted_attrs['%s__%s' % (trackable_model.__name__.lower(),key)] = val
                instance.update( adjusted_attrs )

            final_attrs = {}
            for key,val in instance.iteritems():
                if type(val) == unicode:
                    try:
                        val = str(val)
                    except:
                        continue
                final_attrs[key] = val
            instances.append( final_attrs )

        f = open('./output.csv','w')
        headers = ['id','name',]
        for trackable_model in trackable_models:
            for field in options.get('aggregate_fields').split(','):
                headers.append('%s__%s__%s' % (trackable_model.__name__.lower(),field,'sum'))
                headers.append('%s__%s__%s' % (trackable_model.__name__.lower(),field,'avg'))
        output = csv.DictWriter(f,headers,dialect='excel-tab',quoting=csv.QUOTE_ALL)
        # output = UnicodeWriter(f,instances[0].keys(),quoting=csv.QUOTE_ALL)
        output.writerow( dict(zip(headers,headers)) )
        for instance in instances:
            output.writerow( instance )
        f.close()



"""
This module implements counters as simple wrappers around the current
``MessageBackend``.

"""
from trackable.message import connection


def send_increment_message(request, obj, field_name, data_cls=None, options={}):
    """
    Send a message for incrementing the ``field_name`` count
    for the trackable data model ``data_cls`` by ``value``.
    
    """
    if not options.get('value',None):
        options.update([('value',1)])
    connection.send_message(request, obj, field_name, 'incr', data_cls=data_cls, options=options)

def send_decrement_message(request, obj, field_name, data_cls=None, options={}):
    """
    Send a message for decrementing the ``field_name`` count
    for the trackable data model ``data_cls`` by ``value``.
    
    """
    if not options.get('value',None):
        options.update([('value',1)])
    connection.send_message(request, obj, field_name, 'decr', data_cls=data_cls, options=options)



from django.conf import settings

from celery.registry import tasks
from celery.task import Task

from trackable.message import connection

# COLLECTION_PERIOD = getattr(settings,'TRACKABLE_COLLECTION_PERIOD',1800)
# DISABLE_COLLECTION_TASK = getattr(settings,'TRACKABLE_DISABLE_COLLECTION_TASK',False)


class ProcessTrackableMessages(Task):
    def run(self, **kwargs):
        logger = self.get_logger(**kwargs)
        return connection.process_messages(logger=logger)

tasks.register(ProcessTrackableMessages)



from django.conf.urls.defaults import *

from trackable.tests.models import Page
from trackable.tests.views import page_detail


urlpatterns = patterns('',
    url(r'^page/(?P<slug>[\w-]+)/$', page_detail, {}, name='page-detail'),
)



from django.utils.translation import ugettext_lazy as _
from django.utils import simplejson
from django.core.exceptions import ObjectDoesNotExist
from django.http import ( \
    HttpResponse, HttpResponseRedirect, HttpResponseServerError, Http404)

from trackable import TrackableError
import warnings


def track_object( request, queryset, message_op, message_func, trackable_cls=None,
                  url_field=None, redirect_func=None, 
                  object_id=None, slug=None, slug_field='slug',
                  **options):
    """
    A generic trackable view that aligns a trackable message with the 
    HTTP request/response cycle.

    """
    model = queryset.model
    if object_id:
        queryset = queryset.filter(pk=object_id)
    elif slug and slug_field:
        queryset = queryset.filter(**{slug_field: slug})
    else:
        raise TrackableError("trackable_redirect view must be called with either an object_id or a slug/slug_field.")

    try:
        obj = queryset.get()
    except ObjectDoesNotExist:
        raise Http404("No %s found matching the query" % (model._meta.verbose_name))

    message_func(request,obj,message_op,data_cls=trackable_cls,options=dict(options))

    if request.is_ajax():
        return HttpResponse(simplejson.dumps(True),mimetype='application/javascript')

    if redirect_func:
        return HttpResponseRedirect(redirect_func(obj,dict(options)))
    elif url_field and getattr(obj,url_field,''):
        return HttpResponseRedirect(getattr(obj,url_field))
    elif getattr(obj,'get_absolute_url',''):
        return HttpResponseRedirect(obj.get_absolute_url())

    raise TrackableError( \
        u"url_field has not been specified and object has not defined a get_absolute_url property")

def trackable_redirect( request, queryset, message_op, message_func, trackable_cls=None,
                        url_field=None, redirect_func=None, 
                        object_id=None, slug=None, slug_field='slug',
                        **options):
    warning = DeprecationWarning("`trackable.views.trackable_redirect` has been renamed to `trackable.views.track_object`")
    warnings.warn( warning, stacklevel=2 )    
    return track_object(request,queryset,message_op,message_func,trackable_cls,url_field=url_field,redirect_func=redirect_func,
                        object_id=object_id,slug=slug,slug_field=slug_field,**options)



"""
Emulates the behaviour of a django ``AdminSite``.

Implementation is modified directly from ``django.contrib.admin.sites.AdminSite``
in order to maintain ``AdminSite`` semantics.

"""
from django.utils.translation import ugettext_lazy, ugettext as _
from django.utils.functional import update_wrapper
from django.utils.safestring import mark_safe
from django.db.models.base import ModelBase
from django.utils.text import capfirst
from django.contrib import admin
from django.conf import settings

import re


REGISTER_ADMIN_MODELS = getattr(settings,'TRACKABLE_REGISTER_ADMIN_MODELS',settings.DEBUG)


class AlreadyRegistered(Exception): pass
class NotRegistered(Exception): pass

class TrackableDataAdmin(admin.ModelAdmin):
    list_display = ('parent',)

class TrackableRegistry(object):
    """ """

    def __init__(self, name=None):
        self._registry = {} # model_class class -> admin_class instance

    def register(self, model_or_iterable, trackable_cls=None, admin_cls=None, **options):
        """
        Registers the given model(s) with the given trackable data class.

        The model(s) should be Model classes, not instances.


        If a model is already registered, this will raise AlreadyRegistered.
        """
        if not admin_cls:
            admin_cls = TrackableDataAdmin
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]
        for model in model_or_iterable:
            if model not in self._registry:
                self._registry[model] = [trackable_cls,]
            else:
                self._registry[model].append(
                    trackable_cls
                    )
            if REGISTER_ADMIN_MODELS:
                admin.site.register(trackable_cls,admin_cls)

    def unregister(self, model_or_iterable):
        """
        Unregisters the given model(s).

        If a model isn't already registered, this will raise NotRegistered.
        """
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]
        for model in model_or_iterable:
            if model not in self._registry:
                raise NotRegistered('The model %s is not registered' % model.__name__)
            trackable_cls = self._registry[model]
            del self._registry[model]
            admin.site.unregister(trackable_cls)

    def get_parent(self, trackable_cls):
        """ """
        for model_cls,trackable_models in self._registry.iteritems():
            for trackable_model_cls in trackable_models:
                if trackable_model_cls == trackable_cls:
                    return model_cls

        raise NotRegistered( \
            'The model %s has not been registered to track any model' % trackable_cls.__name__)

# This global object represents the default trackable site, for the common case.
# You can instantiate TrackableRegistry in your own code to create a custom trackable registry.
site = TrackableRegistry()



from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction, IntegrityError
from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes import generic
from django.db import transaction
from django.conf import settings
from django.db.models import F

import datetime


class Spider(models.Model):
    """ """
    user_agent = models.CharField(max_length=128, unique=True)

    class Meta:
        verbose_name = _(u'spider')

class TrackableDataManager(models.Manager):
    """ """
    def get_data_parameters(self, obj, options={}):
        raise NotImplementedError

    def get_filtered_fields(self):
        return ['id','content_type','object_id','parent',]

    def get_trackable_fields(self):
        trackable_fields = []
        for field in self.model._meta.fields:
            if field.name not in self.get_filtered_fields():
                trackable_fields.append(field.name)
        return trackable_fields

    def get_data_object(self, obj, options={}):
        data_obj = None
        params = {}

        try:
            params = self.get_data_parameters(obj,options)
        except NotImplementedError:
            pass
        params.update({
            'content_type':
                ContentType.objects.get_for_model(obj.__class__),
            'object_id':obj.pk
            })

        try:
            return self.get_query_set().get(**params)
        except ObjectDoesNotExist:
            try:
                return self.get_query_set().create(**params)
            except IntegrityError, e:
                transaction.rollback()
                return self.get_query_set().get(**params)
        # except MultipleObjectsReturned:
        #     return self.get_query_set().filter(**params)[0]

class TrackableData(models.Model):
    """ """
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    parent = generic.GenericForeignKey('content_type','object_id')

    class Meta:
        verbose_name_plural = _(u'trackable data')
        abstract = True
        ordering = ('content_type','object_id')
        unique_together = ('content_type','object_id')

    @transaction.commit_on_success
    def op(self, attr, change_func, commit=True):
        try:
            _attr = getattr(self,attr)
        except AttributeError:
            raise TrackableError( \
                u"Attribute %s does not exist" % attr)

        value = change_func( F(attr) )
        setattr(self,attr,value)
        if commit:
            self.save()

        return value

    def incr(self, attr, value=1, commit=True):
        value = long(value)
        return self.op(attr, lambda x:x+value, commit=commit)

    def decr(self, attr, value=1, commit=True):
        value = long(value)
        return self.op(attr, lambda x:x-value, commit=commit)

    def __unicode__(self):
        return u"%s (%s)" % \
            (self._meta.verbose_name,self.parent)

    objects = TrackableDataManager()

class TrackableTimeSeriesDataManager(TrackableDataManager):
    """ """
    def get_filtered_fields(self):
        filtered_fields = super(TrackableTimeSeriesDataManager,self).get_filtered_fields()
        filtered_fields.append('collected_during')
        return filtered_fields

    def get_data_parameters(self, obj, options={}):
        return {
            'collected_during': datetime.date.today(),
            }

class TrackableTimeSeriesData(TrackableData):
    """ """
    collected_during = models.DateField()

    class Meta(TrackableData.Meta):
        verbose_name_plural = _(u'trackable time series data')
        abstract = True
        ordering = ('-collected_during','content_type','object_id')
        unique_together = ('content_type','object_id','collected_during',)

    def __unicode__(self):
        return u"%s (%s) for %s" % \
            (self._meta.verbose_name,self.parent,self.collected_during)

    objects = TrackableTimeSeriesDataManager()



from django.core.management.base import BaseCommand, CommandError
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _
from django.db.models.loading import get_model
from django.db import models, transaction

from optparse import make_option
from trackable.sites import site

import warnings
import sys


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--update-fields', action='store', dest='update_fields',
            help='A series of one or more key:value pairs delimited by commas '
                 'specifying fields from the existing tracking data model '
                 'with which to update the destination trackable data model'), 
        make_option('--key-fields', action='store', dest='key_fields',
            help='Fields when taken together unambiguously identify a single '
                 'record. A series of one or more key:value pairs delimited by commas '
                 'mapping fields from the existing tracking data model to the '
                 'destination trackable data model'), 
        make_option('--link-field', action='store', dest='link_field',
            help='Related model attribute on source_data_model that links to it\'s parent model. Chaining permitted using dot notation, e.g., attribute.other_attribute'),
        make_option('--match-spec', action='store', dest='match_spec',
            help='String option in the form source_attr:app_label.ModelName[:dest_attr]. Use in place of --link-field'),
        make_option('--max-records', action='store', dest='max_records',
            help='Set a limit on the number of records imported from the given source data model'),
        make_option('--order-by', action='store', dest='order_by',
            help='Specify parameters, delimited by commas, to source_queryset.order_by(...)'),

    )
    help = 'Used to import pre-existing tracking data to trackable data stores.'
    args = 'source_data_model dest_trackable_model'

    @transaction.commit_on_success
    def handle_model_data_migration( self, model_cls, trackable_model_cls, update_fields={}, key_fields={}, link_field='parent', match_spec=None, max_records=0, order_by=[] ):
        """ """
        print "Conversion parameters: "
        print "  Source: %s" % (model_cls._meta.verbose_name)
        print "  Destination: %s" % (trackable_model_cls._meta.verbose_name)
        print "  Update fields: %s" % (update_fields)
        print "  Key fields: %s" % (key_fields)
        print "  Link field: %s" % (link_field)
        print "  Match spec: %s" % (match_spec)
        print "  Max records: %d" % (max_records)
        print "  Order by: %s" % (order_by)

        parent_model_cls = site.get_parent( trackable_model_cls )
        source_queryset = model_cls.objects.all()
        if order_by:
            source_queryset = source_queryset.order_by(*order_by)
        if max_records:
            source_queryset = source_queryset[:max_records]

        instance_cnt = source_queryset.count()
        progress_cnt = 0

        for old_instance in source_queryset:

            parent_object_id = None
            if match_spec:
                match_parts = match_spec.split(':')
                if len(match_parts) == 2:
                    source_attr,model_spec = match_parts
                    dest_attr = source_attr
                elif len(match_parts) == 3:
                    source_attr,model_spec,dest_attr = match_parts
                else:
                    raise CommandError("Invalid match_spec option: %s" % match_spec)

                try:
                    parent_model_cls = get_model(*model_spec.split('.'))

                    source_attr_links = source_attr.split('.')
                    source_attr = getattr(old_instance,source_attr_links[0])
                    for link in source_attr_links[1:]:
                        source_attr = getattr(source_attr,link)
                    source_attr_value = source_attr

                    parent_kwargs = {dest_attr:source_attr_value}
                    parent_instance = parent_model_cls.objects.get(**parent_kwargs)
                    parent_object_id = parent_instance.pk
                except AttributeError, e:
                    raise e
                except parent_model_cls.DoesNotExist, e:
                    continue
            else:
                links = link_field.split('.')
                parent_attr = getattr(old_instance,links[0])
                for link in links[1:]:
                    parent_attr = getattr(parent_attr,link)
                parent_object_id = parent_attr.pk

            if not parent_object_id:
                warnings.warn("Could not determine parent for trackable data model. Skipping")
                continue

            filter_kwargs = {}
            filter_kwargs['content_type'] = \
                ContentType.objects.get_for_model(parent_model_cls)
            filter_kwargs['object_id'] = parent_object_id
            filter_kwargs.update( \
                [(field,getattr( old_instance, field )) for field in key_fields] )

            new_instance, created = trackable_model_cls.objects.get_or_create(**filter_kwargs)
            for source_field,dest_field in update_fields.iteritems():
                try:
                    source_field_value = getattr(old_instance,source_field)
                    setattr(new_instance,dest_field,source_field_value)
                except AttributeError:
                    raise CommandError( \
                        "You specified fields in the update option that do not exist.")
            new_instance.save()

            progress_cnt += 1
            if progress_cnt % 100 == 0:
                print ".",
                sys.stdout.flush()

        print "Migrated %d objects" % (progress_cnt)

    def handle(self, *args, **options):
        """ """
        model_spec,trackable_model_spec = (None,None)
        try:
            (model_spec,trackable_model_spec) = args
        except ValueError:
            raise CommandError("Command takes two arguments")

        model_cls = get_model(*model_spec.split('.'))
        trackable_model_cls = get_model(*trackable_model_spec.split('.'))
        kwargs = {}

        if options.get('key_fields',{}):
            kwargs['key_fields'] = dict([ (pair.split(':')[0],pair.split(':')[1]) for pair in options.get('key_fields').split(',') ])
        if options.get('update_fields',{}):
            kwargs['update_fields'] = dict([ (pair.split(':')[0],pair.split(':')[1]) for pair in options.get('update_fields').split(',') ])
        if options.get('link_field',''):
            kwargs['link_field'] = options.get('link_field')
        if options.get('match_spec',''):
            kwargs['match_spec'] = options.get('match_spec')
        if options.get('max_records',0):
            kwargs['max_records'] = long(options.get('max_records'))
        if options.get('order_by',[]):
            kwargs['order_by'] = options.get('order_by').split(',')

        self.handle_model_data_migration( \
            model_cls, trackable_model_cls, **kwargs )



from django.core.management.base import BaseCommand, CommandError
from django.db.models.loading import get_model

from trackable.sites import site, NotRegistered
from trackable.message import connection


class Command(BaseCommand):
    help = 'Used to generate summary data for Vendor tracking data.'
    args = 'app.trackable_model [app.trackable_model ...]'

    def handle(self, *trackable_model_specs, **options):
        for model_spec in trackable_model_specs:
            print "Processing messages for trackable model %s" % (model_spec)
            try:
                app_label,model = model_spec.split('.')
                model_cls = get_model(app_label,model)
                if model_cls is None:
                    raise CommandError("Unknown model: %s.%s" % (app_label,model))
            except ValueError:
                raise CommandError("Use full appname.ModelName specification for argument %s" % model_spec)
            try:
                site.get_parent( model_cls )
            except NotRegistered, e:
                raise CommandError( str(e) )
            cnt = connection.process_messages(model_cls=model_cls)
        if not len(trackable_model_specs):
            print "Processing messages for all trackable models"
            cnt = connection.process_messages()
        print "Processed %d" % (cnt)



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase, TransactionTestCase
from django.conf import settings

from trackable.tests.trackabledata import PageViewTrackableData, PageRevisionTrackableData
from trackable.contrib.celery_tasks import ProcessTrackableMessages
from trackable.message import connection
from trackable.tests.models import Page

from celery.execute import send_task

import multiprocessing
import datetime
import logging
import os.path
import time


class CeleryTest(TransactionTestCase):
    urls = 'trackable.tests.urls'
    fixtures = ['page.json',]
    template_dirs = [
        os.path.join(os.path.dirname(__file__), 'templates'),
    ]

    def setUp(self):
        self.old_template_dir = settings.TEMPLATE_DIRS
        settings.TEMPLATE_DIRS = self.template_dirs

    def tearDown(self):
        settings.TEMPLATE_DIRS = self.old_template_dir

    def testUniprocessingWorkers(self):
        test_page = Page.objects.get(pk=1)
        kwargs = {
            'object_id':test_page.pk,
            'content_type':ContentType.objects.get_for_model(Page)
            }

        self.assertRaises( \
            PageViewTrackableData.DoesNotExist, 
            PageViewTrackableData.objects.get,
            **kwargs
            )

        trackable_data = PageViewTrackableData.objects.get_data_object(test_page)
        self.assertEquals(trackable_data.views,0)

        for i in xrange(200):
            self.client.get(test_page.get_absolute_url())

        trackable_data = PageViewTrackableData.objects.get_data_object(test_page)
        self.assertEquals(trackable_data.views,0)

        t3 = send_task("trackable.contrib.celery_tasks.ProcessTrackableMessages")
        t3.get()

        trackable_data = PageViewTrackableData.objects.get_data_object(test_page)
        self.assertEquals(trackable_data.views,200)


class ConcurrencyTest(TransactionTestCase):
    urls = 'trackable.tests.urls'
    fixtures = ['page.json',]
    template_dirs = [
        os.path.join(os.path.dirname(__file__), 'templates'),
    ]

    def setUp(self):
        self.old_template_dir = settings.TEMPLATE_DIRS
        settings.TEMPLATE_DIRS = self.template_dirs

    def tearDown(self):
        settings.TEMPLATE_DIRS = self.old_template_dir

    # def someTestFunc(self):
    # 
    # ...
    # 
    #     from django.db import connection as dj_connection
    #     dj_connection.close()
    # 
    #     p1 = multiprocessing.Process(target=connection.process_messages)
    #     p2 = multiprocessing.Process(target=connection.process_messages)
    #
    #     p1.start()
    #     p2.start()
    # 
    #     p1.join()
    #     p2.join()
    # 
    # ...

__tests__ = {}



import os.path
import logging


DEBUG = True
TEMPLATE_DEBUG = DEBUG
DEBUG_PROPAGATE_EXCEPTIONS = DEBUG
THUMBNAIL_DEBUG = DEBUG

ADMINS = (
    ('Thom Linton', 'thom.linton@gmail.com'),
)
MANAGERS = ()
INTERNAL_IPS = (
    '127.0.0.1',
)

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/Los_Angeles'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = False
USE_ETAGS = False

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
    'django.template.loaders.eggs.load_template_source',
)

# default timeout is 300s 
CACHE_BACKEND = 'memcached://127.0.0.1:11211/'
# CACHE_BACKEND = 'newcache://127.0.0.1:11211/?binary=true'

# CACHE_MIDDLEWARE_SECONDS = 300
# CACHE_MIDDLEWARE_ANONYMOUS_ONLY = True

MIDDLEWARE_CLASSES = [
    'django.middleware.common.CommonMiddleware',
    'django.middleware.transaction.TransactionMiddleware',
]

TEMPLATE_CONTEXT_PROCESSORS = [
    "django.core.context_processors.debug",
]

AUTHENTICATION_BACKENDS = []

INSTALLED_APPS = [
    'django.contrib.contenttypes',

    'trackable',
    'trackable.tests',

    'south',
    'djcelery',
]

ROOT_URLCONF = 'tests.test_urls'

# Base path info
BASE_PATH = os.path.dirname(__file__)

# Testing settings
# TEST_RUNNER = "djcelery.contrib.test_runner.run_tests"

# Celery settings
# CELERY_RESULT_BACKEND = "cache"
CELERY_QUEUES = {
    "trackable": {
        "exchange": "direct",
        "binding_key": "trackable",
        },
}
CELERY_DEFAULT_QUEUE = "trackable"
CELERY_DEFAULT_EXCHANGE_TYPE = "direct"
CELERY_DEFAULT_ROUTING_KEY = "trackable"

# CELERY_SEND_TASK_ERROR_EMAILS = False
# CELERY_SEND_EVENTS = True

# CELERYD_CONCURRENCY = 2
CELERYD_LOG_LEVEL = 'INFO'

# Carrot settings
BROKER_HOST = "localhost"
BROKER_PORT = 5672
BROKER_VHOST = "/"
BROKER_USER = "guest"
BROKER_PASSWORD = "guest"

TRACKABLE_USER_AGENT_FILTERING = False
TRACKABLE_REMOVE_MALFORMED_MESSAGES = True
TRACKABLE_CAPTURE_CONNECTION_ERRORS = True
TRACKABLE_DISABLE_COLLECTION_TASK = True
# TRACKABLE_LOGLEVEL = logging.DEBUG
TRACKABLE_ENGINE = 'trackable.message.backends.carrot'

FIXTURE_DIRS = (
    os.path.join( BASE_PATH, 'fixtures' )
)
TEMPLATE_DIRS = (
    os.path.join( BASE_PATH, 'templates')
)

try:
    from local_settings import *

    import djcelery
    djcelery.setup_loader()

except ImportError:
    pass



### these models are for testing

from django.utils.translation import ugettext_lazy as _
from django.db import models


class Page(models.Model):
    """ """
    title = models.CharField(_(u'title'), max_length=128)
    slug = models.SlugField(_(u'slug'))
    abstract = models.TextField(_(u'abstract'))
    content = models.TextField(_(u'content'))

    published_on = models.DateTimeField(_(u'published on'))

    class Meta:
        ordering = ('published_on',)

    def get_absolute_url(self):
        return ('page-detail', (), {'slug':self.slug})
    get_absolute_url = models.permalink(get_absolute_url)

    def __unicode__(self):
        return u"%s" % self.title



from django.core.mail import mail_admins
from django.conf import settings

from carrot.connection import DjangoBrokerConnection
from carrot.messaging import Publisher, Consumer

from trackable.message.backends.base import BaseMessageBackend


class MessageBackend(BaseMessageBackend):
    """ """
    def __init__(self):
        try:
            self.connection = DjangoBrokerConnection()
            self.consumer = Consumer( \
                connection=self.connection,
                queue='trackable_internal',
                exchange='direct',
                routing_key='trackable_internal',
                )
            self.publisher = Publisher( \
                connection=self.connection,
                queue='trackable_internal',
                exchange='direct',
                routing_key='trackable_internal',
                serializer='pickle',
                )
        except Exception, e:
            if settings.DEBUG:
                raise e
            mail_admins(
                '[django-trackable] Error while acquiring a connection in carrot backend',
                '%s' % (str(e))
                )


    def send(self, payload):
        try:
            return self.publisher.send( payload )
        except Exception, e:
            if settings.DEBUG:
                raise e
            mail_admins(
                '[django-trackable] Error while sending message in carrot backend',
                '%s' % (str(e))
                )

    def recv(self):
        try:
            return self.consumer.fetch()
        except Exception, e:
            if settings.DEBUG:
                raise e
            mail_admins(
                '[django-trackable] Error while sending message in carrot backend',
                '%s' % (str(e))
                )



# from trackable.models import TrackableData, TrackableTimeSeriesData, TrackableError
# from trackable.sites import TrackableRegistry, site


class TrackableError(Exception):
    pass

def autodiscover():
    """
    Auto-discover INSTALLED_APPS trackabledata.py modules and fail silently when
    not present. 

    """
    import copy
    from django.conf import settings
    from django.utils.importlib import import_module
    from django.utils.module_loading import module_has_submodule

    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        # Attempt to import the app's admin module.
        try:
            before_import_registry = copy.copy(site._registry)
            import_module('%s.trackabledata' % app)
        except:
            # Reset the model registry to the state before the last import as
            # this import will have to reoccur on the next request and this
            # could raise NotRegistered and AlreadyRegistered exceptions
            # (see #8245).
            site._registry = before_import_registry

            # Decide whether to bubble up this error. If the app just
            # doesn't have an admin module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, 'trackabledata'):
                raise




from setuptools import setup, find_packages
import os

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "django-trackable",
    version = "0.3.5",
    author = "Thom Linton",
    author_email = "thom.linton@gmail.com",
    description = "A Django application which provides tools for out-of-band capture of arbitrary tracking data.",
    license = "BSD",
    keywords = "django tracking",
    url = "http://github.com/yorkedork/django-trackable/tree/master",
    packages=find_packages(),
    install_requires = [
        'django-celery',
        ],
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: BSD License",
    ],
    package_data = {
        'trackable': ['fixtures/*.json',],
        'trackable.tests': ['fixtures/*.json','templates/*.html',],
        },
)



from django.utils.translation import ugettext_lazy as _
from django.db import models

from trackable.models import TrackableData, TrackableTimeSeriesData
from trackable.tests.models import Page
from trackable.sites import site


class PageViewTrackableData(TrackableData):
    """ """
    views = models.PositiveIntegerField(_(u'page views'), default=0)

    class Meta(TrackableData.Meta):
        verbose_name_plural = _(u'page views')

    def __unicode__(self):
        return u"%s (%s views)" % (self.parent,self.views)

class PageRevisionTrackableData(TrackableTimeSeriesData):
    """ """
    revisions = models.PositiveIntegerField(_(u'page revisions'), default=0)

    class Meta(TrackableTimeSeriesData.Meta):
        verbose_name_plural = _(u'page revisions')

    def __unicode__(self):
        return u"%s (%s revisions)" % (self.parent,self.revisions)

# trackable.site.register(Page,PageViewTrackableData)
site.register(Page,PageViewTrackableData)
# trackable.site.register(Page,PageRevisionTrackableData)
site.register(Page,PageRevisionTrackableData)
